(* icosactlr.ml     Controller for IcosaBlue example
 *
 * Copyright (c) 2011 Psellos   http://psellos.com
 * Licensed under the MIT license:
 *     http://www.opensource.org/licenses/mit-license.php
 *)

let facect = 20    (* Number of faces of icosahedron *)
let sidect = 3     (* Number of sides of each face (triangle) *)

let txfullwd = 512 (* Full texture width (two triangle images) *)
let txwd = 256     (* Texture width (one triangle) *)
let txht = 256     (* Texture height *)

let ftxfullwd = float_of_int txfullwd
let ftxwd = float_of_int txwd
let ftxht = float_of_int txht

let tpowd = 512    (* Teapot background texture width *)
let tpoht = 512    (* Teapot background texture height *)

let ftpowd = float_of_int tpowd
let ftpoht = float_of_int tpoht

(* Geometry of icosahedron texture image.
 *)
let s3o2 = sqrt 3.0 /. 2.0   (* Equilateral triangle constant *)
let imgside = 189.0          (* Triangle side *)
let imgheight =              (* Triangle height *)
    Cocoa.frnd (imgside *. s3o2)
let inmar = 20.0             (* Margin of inner triangle *)

(* Colors.
 *)
let clear = (0.0, 0.0, 0.0, 0.0)

let dark_blue = (0.153, 0.263, 0.733, 1.0)
let bright_blue = (0.125, 0.196, 0.918, 1.0)
let beige = (0.882, 0.824, 0.796, 1.0)

let dark_orange = (0.757, 0.376, 0.227, 1.0)
let light_orange = (0.808, 0.412, 0.259, 1.0)


let draw_triangle cgctx xorigin =
    (* Draw the triangular texture image for a side of the icosahedron.
     * There's a background color and an inner triangle with rounded
     * corners.
     *)
    let s = imgside  (* Convenient abbreviation of image side*)
    in let i = inmar (* Convenient abbreviation of indent *)
    in let t = 0.1   (* Trimmed distance, beginning of rounded corner *)
    in let c = 0.05  (* Control distance, position of Bezier control point *)
    in let segs = [|
        (* Segments making up the inner triangle.
         *)
        ((i *. s3o2, i /. 2.0), (s /. 2.0, s *. s3o2 -. i));
        ((s /. 2.0, s *. s3o2 -. i), (s -. i *. s3o2, i /. 2.0));
        ((s -. i *. s3o2, i /. 2.0), (i *. s3o2, i /. 2.0));
    |]
    in let otrans ((x1, y1), (x2, y2)) =
        ((xorigin +. x1, y1), (xorigin +. x2, y2))
    in let osegs = Array.map otrans segs
    in let trimseg f ((x1, y1), (x2, y2)) =
        ((x1 +. f *. (x2 -. x1),
          y1 +. f *. (y2 -. y1)),
         (x1 +. (1.0 -. f) *. (x2 -. x1),
         y1 +. (1.0 -. f) *. (y2 -. y1)))
    in let tsegs = Array.map (trimseg t) osegs
    in let csegs = Array.map (trimseg c) osegs
    in let add_side (_, p) (_, c1) (c2, _) (q, _) =
        begin
        CgContext.addLineToPoint cgctx p;
        CgContext.addCurveToPoint cgctx c1 c2 q;
        end
    in
        begin
        CgContext.setFillColor cgctx dark_blue;
        CgContext.fillRect cgctx (xorigin, 0.0, ftxwd, ftxht);
        CgContext.moveToPoint cgctx (fst tsegs.(0));
        for i = 0 to 2 do
            let i' = (i + 1) mod 3
            in
                add_side tsegs.(i) csegs.(i) csegs.(i') tsegs.(i')
        done;
        CgContext.setFillColor cgctx bright_blue;
        CgContext.fillPath cgctx;
        end


let text_width cgctx text =
    (* Calculate the width of the text using current settings of the
     * CgContext.  Caller must beware that the text drawing mode will be
     * changed by this call.
     *)
    let () =
        begin
        CgContext.setTextPosition cgctx (0.0, 0.0);
        CgContext.setTextDrawingMode cgctx CgContext.Invisible;
        CgContext.showText cgctx text (String.length text);
        end
    in
        fst (CgContext.getTextPosition cgctx)


let draw_psellos cgctx xorigin =
    (* Draw a Greek Psi and Psellos.com into the triangle for "branding"
     * purposes.  Forgive the self-promotion.
     *)
    let psimar = 17.0   (* Extra margin around the Psi *)
    in let txbase = 8.0 (* Baseline for text *)
    in let text = "Psellos.com"
    in let rect =
        (xorigin +. inmar +. psimar,
         inmar +. psimar,
         imgside -. 2.0 *. inmar -. 2.0 *. psimar,
         imgheight -. 3.0 *. inmar -. 2.0 *. psimar)
    in let sfig = Bzpdata.bzp_scale Bzpdata.figure1 rect
    in let drawit = function
        | Bzpdata.BZBegin -> CgContext.beginPath cgctx
        | Bzpdata.BZMove p -> CgContext.moveToPoint cgctx p
        | Bzpdata.BZLine p -> CgContext.addLineToPoint cgctx p
        | Bzpdata.BZCurve (c, d, p) -> CgContext.addCurveToPoint cgctx c d p
        | Bzpdata.BZClose -> CgContext.closePath cgctx
    in
        begin
        (* First the Greek Psi.
         *)
        Bzpdata.bzp_iter drawit sfig;
        CgContext.setFillColor cgctx beige;
        CgContext.fillPath cgctx;

        (* Now the text.
         *)
        CgContext.selectFont cgctx "Verdana" 16.0;
        let indent = floor ((imgside -. text_width cgctx text) /. 2.0)
        in let pos = (xorigin +. indent, inmar +. txbase)
        in
            CgContext.setTextPosition cgctx pos;
        CgContext.setTextDrawingMode cgctx CgContext.FillStroke;
        CgContext.setFillColor cgctx beige;
        CgContext.setStrokeColor cgctx beige;
        CgContext.showText cgctx text (String.length text);
        end


let select_texture faceix txix =
    (* Select the given subtexture for the given face.  Right now there
     * are only two subtextures, numbered 0 and 1.  The default is 0,
     * which is plain.  Subtexture 1 has the "branding".
     *)
    let xorigin = (float_of_int txix *. ftxwd) /. ftxfullwd
    in
        for i = 0 to 2 do
            let vix = Raw.get Icosamodel.faces ~pos: (3 * faceix + i)
            in let xold = Raw.get_float Icosamodel.texcos ~pos: (2 * vix)
            in let xoffset = mod_float xold (ftxwd /. ftxfullwd)
            in
                Raw.set_float Icosamodel.texcos ~pos: (2 * vix)
                    (xorigin +. xoffset)
        done


let draw_psi cgctx =
    (* Draw just a Greek Psi.
     *)
    let rect = (0.0, 0.0, ftpowd, ftpoht)
    in let sfig = Bzpdata.bzp_scale Bzpdata.figure1 rect
    in let drawit = function
        | Bzpdata.BZBegin -> CgContext.beginPath cgctx
        | Bzpdata.BZMove p -> CgContext.moveToPoint cgctx p
        | Bzpdata.BZLine p -> CgContext.addLineToPoint cgctx p
        | Bzpdata.BZCurve (c, d, p) -> CgContext.addCurveToPoint cgctx c d p
        | Bzpdata.BZClose -> CgContext.closePath cgctx
    in
        begin
        CgContext.setFillColor cgctx dark_orange;
        CgContext.fillRect cgctx rect;
        Bzpdata.bzp_iter drawit sfig;
        CgContext.setFillColor cgctx light_orange;
        CgContext.fillPath cgctx;
        end


class t =
object (self)
    inherit Wrappee.t

    val interval = 1.0 /. 24.0  (* Interval between timer ticks *)

    val mutable active : bool = true
    val mutable ticks = 0       (* Number of ticks since inception *)

    (* View that delegates to us.
     *)
    val mutable theDelegator: UiView.t option = None

    (* OpenGL state.
     *)
    val mutable theGlctx: EAGLContext.t option = None
    val mutable view_renderbuffer = GlFramebuffer.renderbuffer0
    val mutable view_depthbuffer = GlFramebuffer.renderbuffer0
    val mutable view_framebuffer = GlFramebuffer.framebuffer0
    val mutable backing_width = 0
    val mutable backing_height = 0

    (* Textures.
     *)
    val mutable theCgctxs : CgContext.t list = []
    val mutable thePixmaps : ([`rgba], [`ubyte]) GlPix.t list = []
    val mutable theTexids : GlTex.texture_id list = []
    val mutable curTexid = -1 (* Index of current texture id; -1 -> uninit *)

    (* Lighting
     *)
    val mutable curLighting = -1 (* Current lighting setup; -1 -> uninit *)

    (* Accessors.
     *)
    method delegator =
        match theDelegator with
        | None -> failwith "Initialization error: no delegator (IcosaBlue.nib)"
        | Some view -> view
    
    method setDelegator' view =
        theDelegator <- Some view

    method glctx =
        match theGlctx with
        | None -> failwith "Initialization error: no OpenGL context"
        | Some c -> c

    method set_glctx glctx =
        theGlctx <- Some glctx

    method cgctx n =
        if List.length theCgctxs <= n then
            failwith "Initialization error: too few CgContexts"
        else
            List.nth theCgctxs n

    method set_cgctxs cs =
        theCgctxs <- cs

    method pixmap n =
        if List.length thePixmaps <= n then
            failwith "Initialization error: no GlPix"
        else
            List.nth thePixmaps n

    method set_pixmaps ps =
        thePixmaps <- ps

    method texid n =
        if List.length theTexids <= n then
            failwith "Initialization error: too few texture ids"
        else
            List.nth theTexids n

    method set_texids tids =
        theTexids <- tids

    method delete_texids =
        begin
        if theTexids <> [] then
            GlTex.delete_textures (Array.of_list theTexids);
        theTexids <- [];
        end

    method texbind n =
        if curTexid <> n then
            begin
            GlTex.bind_texture ~target: `texture_2d (self#texid n);
            curTexid <- n;
            end

    (* Application state change events.
     *)
    method applicationDidFinishLaunching' (app: UiApplication.t) =
        (* Startup code.
         *)
        begin
        (UiApplication.sharedApplication ())#setStatusBarStyle'
            UiApplication.StatusBarStyleBlackOpaque;
        self#delegator#layer#setOpaque' true;
        self#set_glctx
            ((EAGLContext.alloc ())#initWithAPI'
                    EAGLContext.RenderingAPIOpenGLES1);
        self#set_cgctxs
            [CgContext.create txfullwd txht;
             CgContext.create tpowd tpoht];
        ignore (
            NsTimer.scheduledTimer5 interval self "drawView'" Wrappee.nil true
        );

        (* Set coordinate systems of texture graphics contexts to
         * match the OpenGL coordinate system.
         *)
        CgContext.scaleCTM (self#cgctx 0) 1.0 (-1.0);
        CgContext.translateCTM (self#cgctx 0) 0.0 (-. ftxht);
        CgContext.setTextMatrix (self#cgctx 0)
            (1.0, 0.0, 0.0, -1.0, 0.0, 0.0);

        (* Draw faces of icosahedron and select which ones are where.
         *)
        draw_triangle (self#cgctx 0) 0.0;
        draw_triangle (self#cgctx 0) ftxwd;
        draw_psellos (self#cgctx 0) ftxwd;

        select_texture 4 1;  (* Just one face has the "branding". *)

        (* Draw background for the teapot.
         *)
        draw_psi (self#cgctx 1);

        (* Make pixmaps for textures.
         *)
        self#set_pixmaps
            [GlPix.of_raw (CgContext.getData (self#cgctx 0))
                ~format: `rgba ~width: txfullwd ~height: txht;
             GlPix.of_raw (CgContext.getData (self#cgctx 1))
                ~format: `rgba ~width: tpowd ~height: tpoht];
        end

    method applicationDidEnterBackground' (app: UiApplication.t) =
        active <- false

    method applicationWillEnterForeground' (app: UiApplication.t) =
        active <- true

    method applicationDidReceiveMemoryWarning' (app: UiApplication.t) =
        Gc.compact () (* Best effort to reclaim space *)

    (* OpenGL drawing.
     *)

    method orthographic =
        (* Orthographic projection in pixel coordinates.
         *)
        let fw = float_of_int backing_width
        in let fh = float_of_int backing_height
        in
            begin
            GlMat.mode `projection;
            GlMat.load_identity ();
            GlMat.ortho
                ~x: (0.0, fw)
                ~y: (0.0, fh)
                ~z: (-. fw /. 2.0, fw /. 2.0);
            GlMat.mode `modelview;
            end

    method perspective =
        (* Perspective projection.
         *)
        begin
        GlMat.mode `projection;
        GlMat.load_identity ();
        let near, far, fieldofview = 0.01, 1000.0, atan 1.0
        in let size = near *. tan (fieldofview /. 2.0)
        in let rat = float_of_int backing_height /. float_of_int backing_width
        in
            GlMat.frustum
                ~x: (-. size, size)
                ~y: (-. size *. rat, size *. rat)
                ~z: (near, far);
        GlMat.mode `modelview;
        end

    method lighting n =
        if curLighting <> n then
            match n with
            | 0 -> (* Weekday *)
                begin
                GlLight.light 0 (`ambient (0.3, 0.3, 0.3, 1.0));
                GlLight.light 0 (`diffuse (0.6, 0.6, 0.6, 1.0));
                GlLight.light 0 (`specular (0.5, 0.5, 0.5, 1.0));
                GlLight.light 0 (`position (10.0, 10.0, 10.0, 1.0));
                GlLight.light 0 (`spot_direction (-1.0, -1.0, -1.0));
                GlLight.light 0 (`spot_cutoff 20.0);

                GlLight.material ~face: `both
                    (`ambient_and_diffuse (1.0, 1.0, 1.0, 1.0));
                GlLight.material ~face: `both
                    (`specular (0.3, 0.3, 0.3, 1.0));
                GlLight.material ~face: `both
                    (`shininess 30.0);
                curLighting <- 0;
                end
            | 1 -> (* Weekend *)
                begin
                GlLight.light 0 (`ambient (0.3, 0.3, 0.3, 1.0));
                GlLight.light 0 (`diffuse (0.6, 0.6, 0.6, 1.0));
                GlLight.light 0 (`specular (0.8, 0.8, 0.8, 1.0));
                GlLight.light 0 (`position (10.0, 10.0, 10.0, 1.0));
                GlLight.light 0 (`spot_direction (-1.0, -1.0, -1.0));
                GlLight.light 0 (`spot_cutoff 20.0);

                GlLight.material ~face: `both
                    (`ambient_and_diffuse (0.341, 0.451, 0.639, 1.0));
                GlLight.material ~face: `both
                    (`specular (1.0, 1.0, 1.0, 1.0));
                GlLight.material ~face: `both
                    (`shininess 20.0);
                curLighting <- 1;
                end
            | _ -> ()
            
    method viewLayoutSubviews' (view: UiView.t) =
        (* This method is called at the beginning, and after changes to
         * the geometry of the view.  (Re)allocate and (re)initialize
         * the framebuffer state.
         *)
        begin
        ignore self#make_current;
        self#destroyFramebuffer;
        ignore self#createFramebuffer;

        Gl.enable `depth_test;

        (* Texture state.
         *)
        Gl.enable `texture_2d;
        GlTex.env (`mode `modulate);
        self#delete_texids;
        self#set_texids (Array.to_list (GlTex.gen_textures 2));
        for i = 0 to 1 do
            self#texbind i;
            List.iter (GlTex.parameter ~target: `texture_2d) [
                (`wrap_s `clamp_to_edge);
                (`wrap_t `clamp_to_edge);
                (`mag_filter `linear);
                (`min_filter `linear);
            ];
            GlTex.image2d ~internal: `rgba (self#pixmap i);
        done;

        GlDraw.viewport ~x: 0 ~y: 0 ~w: backing_width ~h: backing_height;

        (* Enable lighting.
         *)
        Gl.enable `lighting;
        Gl.enable `light0;

        (* Model/view matrix is continually updated to do animation.
         *)
        GlMat.mode `modelview;

        self#drawView' (Obj.magic Wrappee.nil);
        end

    method drawView' (timer: NsTimer.t) =
        (* Display one frame.
         *)
        if active then
            begin
            ignore self#make_current;
            GlFramebuffer.bind_framebuffer `framebuffer view_framebuffer;
            match (Unix.localtime (Unix.time ())).Unix.tm_wday with
            | 0 | 6 -> self#draw_weekend_view
            | _ -> self#draw_weekday_view
            end

    method draw_weekday_view =
        (* Display one frame of the icosahedron.
         *)
        begin
        (* Each frame rotates a little and displays the new view.
         * To keep things slightly more interesting, the axis of
         * rotation moves in a circle.
         *)
        self#perspective;
        GlMat.load_identity ();
        GlMat.translate ~z: (-3.0) ();
        let angle = -2.0 *. float_of_int ticks
        in let precession = float_of_int ticks *. 0.00314159265358979324
        in let inclination = 0.2
        in let x = inclination *. cos precession
        in let z = inclination *. sin precession
        in
            GlMat.rotate ~angle ~x ~y: 1.0 ~z ();
        GlMat.rotate ~angle: 58.282525588538995 ~x: 1.0 ();

        ticks <- ticks + 1;

        self#texbind 0;
        self#lighting 0;

        GlClear.color (0.0, 0.0, 0.0);
        GlClear.clear [ `color ; `depth ];
        GlArray.enable `vertex;
        GlArray.enable `normal;
        GlArray.enable `texture_coord;
        GlArray.vertex `three Icosamodel.vertices;
        GlArray.normal Icosamodel.normals;
        GlArray.tex_coord `two Icosamodel.texcos;
        GlArray.draw_elements
            `triangles ~count: (sidect * facect) Icosamodel.faces;
        GlArray.disable `vertex;
        GlArray.disable `normal;
        GlArray.disable `texture_coord;
        
        GlFramebuffer.bind_renderbuffer `renderbuffer view_renderbuffer;
        self#present_buffer;
        end

    method draw_background =
        (* Draw a background, always behind everything else.  The
         * background picture is in texture 1.
         *)
        begin
        Gl.disable `lighting;
        self#orthographic;
        GlMat.load_identity ();
        self#texbind 1;
        let create fs =
            let res = Raw.create_static `float (Array.length fs)
            in let () = Array.iteri (fun i v -> Raw.set_float res i v) fs
            in
                res
        in let tbuf = create [|
            0.0; 0.0;   0.0; 1.0;   1.0; 1.0;   1.0; 0.0
        |]
        in let fw = float_of_int backing_width
        in let fh = float_of_int backing_height
        in let far = -. fw /. 2.0 +. 1.0
        in let (x, y, w, h) =
            if fw > fh then
                let m = fw /. 25.0
                in
                   ((fw -. fh) /. 2.0 +. m, m, fh -. 2.0 *. m, fh -. 2.0 *. m)
            else
                let m = fh /. 25.0
                in
                    (m, (fh -. fw) /. 2.0 +. m, fw -. 2.0 *.m, fw -. 2.0 *. m)
        in let vbuf = create [|
            x; y; far;  x; y +. h; far;  x +. w; y +. h; far;  x +. w; y; far;
        |]
        in
            begin
            Gl.enable `texture_2d;
            GlArray.enable `texture_coord;
            GlArray.tex_coord `two tbuf;
            GlArray.enable `vertex;
            GlArray.vertex `three vbuf;

            GlArray.draw_arrays `triangle_fan 0 4;

            GlArray.disable `vertex;
            GlArray.disable `texture_coord;
            Gl.disable `texture_2d;
            Raw.free_static vbuf;
            Raw.free_static tbuf;
            end;
        Gl.enable `lighting;
        end

    method draw_weekend_view =
        (* Display one frame of the teapot.
         *)
        begin
        let (r, g, b, a) = dark_orange
        in
            GlClear.color ~alpha: a (r, g, b);
        GlClear.clear [ `color ];

        self#draw_background;

        self#perspective;
        self#lighting 1;
        GlClear.clear [ `depth ];
        GlMat.load_identity ();
        GlMat.translate ~x: 0.0 ~y: (-0.25) ~z: (-1.1) ();
        GlMat.scale ~x: 2.5 ~y: 2.5 ~z: 2.5 ();
        GlMat.translate ~y: 0.04 ();
        GlMat.rotate ~angle: 30.0 ~x: 1.0 ~y: 0.0 ~z: 0.0 ();

        let angle = -2.0 *. float_of_int ticks
        in
            GlMat.rotate ~angle ~x: 0.0 ~y: 1.0 ~z: 0.0 ();

        ticks <- ticks + 1;

        GlArray.enable `vertex;
        GlArray.enable `normal;
        GlArray.vertex `three Teapotmodel.vertices;
        GlArray.normal Teapotmodel.normals;
        let rec drawelems ix =
            if ix < Raw.length Teapotmodel.newindices then
                let len = Raw.get ~pos: ix Teapotmodel.newindices
                in let slice = Raw.sub Teapotmodel.newindices (ix + 1) len
                in let () =
                    GlArray.draw_elements `triangle_strip ~count: len slice
                in
                    drawelems (ix + len + 1)
        in
            drawelems 0;
        GlArray.disable `vertex;
        GlArray.disable `normal;
        
        GlFramebuffer.bind_renderbuffer `renderbuffer view_renderbuffer;
        self#present_buffer;
        end

    method private make_current : bool =
        EAGLContext.setCurrentContext' self#glctx

    method private buffer_storage : bool =
        self#glctx#renderbufferStorage'fromDrawable'
            `renderbuffer self#delegator#layer

    method private present_buffer =
        ignore (self#glctx#presentRenderbuffer' `renderbuffer)
    
    method private createFramebuffer =
        let () =
            begin
            view_framebuffer <- GlFramebuffer.gen_framebuffer ();
            view_renderbuffer <- GlFramebuffer.gen_renderbuffer ();
            GlFramebuffer.bind_framebuffer `framebuffer view_framebuffer;
            GlFramebuffer.bind_renderbuffer `renderbuffer view_renderbuffer;
            ignore self#buffer_storage;
            GlFramebuffer.renderbuffer
                `framebuffer `color0 `renderbuffer view_renderbuffer;
            backing_width <-
                GlFramebuffer.get_renderbuffer_parameter `renderbuffer `width;
            backing_height <-
                GlFramebuffer.get_renderbuffer_parameter `renderbuffer `height;
            (* Depth buffer
             *)
            view_depthbuffer <- GlFramebuffer.gen_renderbuffer ();
            GlFramebuffer.bind_renderbuffer `renderbuffer view_depthbuffer;
            GlFramebuffer.renderbuffer_storage
                `renderbuffer `depth_component16 backing_width backing_height;
            GlFramebuffer.renderbuffer
                `framebuffer `depth `renderbuffer view_depthbuffer;
            end
        in
            GlFramebuffer.check_status `framebuffer = `complete
        
    method private destroyFramebuffer =
        begin
        GlFramebuffer.delete_framebuffer view_framebuffer;
        view_framebuffer <- GlFramebuffer.framebuffer0;
        GlFramebuffer.delete_renderbuffer view_renderbuffer;
        view_renderbuffer <- GlFramebuffer.renderbuffer0;
        GlFramebuffer.delete_renderbuffer view_depthbuffer;
        view_depthbuffer <- GlFramebuffer.renderbuffer0;
        end
end

let wrapped robjcv =
    let c = new t in c#setContainer robjcv; c

let () = Callback.register "Icosactlr.wrapped" wrapped
